/*
 * Simple synchronous userspace interface to SPI devices
 *
 * Copyright (C) 2006 SWAPP
 *	Andrea Paterniani <a.paterniani@swapp-eng.it>
 * Copyright (C) 2007 David Brownell (simplification, cleanup)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/ioctl.h>
#include <linux/fs.h>
#include <linux/device.h>
#include <linux/err.h>
#include <linux/list.h>
#include <linux/errno.h>
#include <linux/mutex.h>
#include <linux/slab.h>
#include <linux/compat.h>
#include <linux/of.h>
#include <linux/of_device.h>
#include <linux/acpi.h>

#include <linux/spi/spi.h>
#include <linux/spi/spidev.h>

#include <linux/uaccess.h>
#include <linux/gpio/consumer.h>
#include <linux/delay.h>
#include <linux/of_gpio.h>
/*-------------------------------------------------------------------------*/

static struct spi_device *rc522_spi_device;
static struct device *rc522_device;
static int major;
static struct gpio_desc *rst_gpio;
static struct gpio_desc *cs_gpio;

//static struct device_node *rc522_node;

static unsigned char read_one_reg(unsigned char reg);
static void write_one_reg(unsigned char reg, unsigned char value);

static void rc522_reset_disable(void)
{
	gpiod_set_value(rst_gpio, 1);
}

static void rc522_reset_enable(void)
{
	gpiod_set_value(rst_gpio, 0);
}

static int rc522_drv_open(struct inode *inode, struct file *file)
{
	printk("%s %s line %d\n", __FILE__, __FUNCTION__, __LINE__);
	rc522_reset_disable();
	udelay(1);
	rc522_reset_enable();
	udelay(1);
	rc522_reset_disable();
	udelay(1);
	return 0;
}

static ssize_t rc522_drv_read(struct file *file, char __user *buf, size_t size, loff_t *offset)
{
	int ret = 0;
	unsigned char ker_buf[2];
	memset(ker_buf, 0, sizeof(ker_buf) / sizeof(ker_buf[0]));
	
	ret = copy_from_user(&ker_buf[0], buf, 1);
	if(ret < 0)
	{
		printk("copy_from_user err!\n");
		return -1;
	}
	
	ker_buf[1] = read_one_reg(ker_buf[0]);
	ret = copy_to_user((void *)buf, &ker_buf[1], 1);
	if(ret < 0)
	{
		printk("copy_to_user err!\n");
		return -1;
	}
	return 0;
}

static ssize_t rc522_drv_write(struct file *file, const char __user *buf, size_t size, loff_t *offset)
{
	int ret = 0;
	unsigned char ker_buf[2];
	memset(ker_buf, 0, sizeof(ker_buf) / sizeof(ker_buf[0]));
	
	ret = copy_from_user(ker_buf, buf, 2);
	if(ret < 0)
	{
		printk("copy_from_user err!\n");
		return -1;
	}
	write_one_reg(ker_buf[0], ker_buf[1]);
	return 0;
}

static int rc522_drv_close(struct inode *inode, struct file *file)
{
	printk("%s %s line %d\n", __FILE__, __FUNCTION__, __LINE__);
	rc522_reset_enable();
	gpiod_put(rst_gpio);
	gpiod_put(cs_gpio);
	return 0;
}


static const struct file_operations spidev_fops = {
	.owner =	THIS_MODULE,
	/* REVISIT switch to aio primitives, so that userspace
	 * gets more complete API coverage.  It'll simplify things
	 * too, except for the locking.
	 */
	.open    = rc522_drv_open,
	.read    = rc522_drv_read,
	.write   = rc522_drv_write,
	.release = rc522_drv_close,
};
int value;

static int spi_read_regs(unsigned char reg, unsigned char *buf, int len)
{
	int ret = 0;
	unsigned char txdata[len];
	struct spi_message message;
	struct spi_transfer *transfer;

	/* 片选使能 */
	gpiod_set_value(cs_gpio, 0);

	/* 申请并初始化 spi_transfer */
	transfer = kzalloc(sizeof(struct spi_transfer), GFP_KERNEL);

	/* 将读寄存器地址修改为正确的格式 addr = ((reg << 1) & 0x7E) | 0x80 */
	txdata[0] = ((reg << 1) & 0x7E) | 0x80;
	transfer->tx_buf = txdata;
	transfer->len = 1;

	/* 初始化 spi_message */
	spi_message_init(&message);

	/* 将 spi_transger 加入 spi_message 的尾部 */
	spi_message_add_tail(transfer, &message);

	/* 开启 spi_sync 传输地址 */
	ret = spi_sync(rc522_spi_device, &message);

	if(ret < 0)
	{
		printk("spi_sync err!\n");
		return -1;
	}

	/* 读该寄存器的数据 */
	txdata[0] = 0xff;
	transfer->rx_buf = buf;
	transfer->len = len;

	/* 初始化 spi_message */
	spi_message_init(&message);

	/* 将 spi_transger 加入 spi_message 的尾部 */
	spi_message_add_tail(transfer, &message);

	/* 开启 spi_sync 接收数据 */
	ret = spi_sync(rc522_spi_device, &message);
	if(ret < 0)
	{
		printk("spi_sync err!\n");
		return -1;
	}

	/*释放 spi_transfer */
	kfree(transfer);

	/* 片选失能 */
	gpiod_set_value(cs_gpio, 1);
	
	return ret;
}

static int spi_write_regs(unsigned char reg, unsigned char *buf, int len)
{
	int ret = 0;
	unsigned char txdata[len];
	struct spi_message message;
	struct spi_transfer *transfer;
	
	/* 片选使能 */
	gpiod_set_value(cs_gpio, 0);

	/* 申请并初始化 spi_transfer */
	transfer = kzalloc(sizeof(struct spi_transfer), GFP_KERNEL);

	/* 将写寄存器地址修改为正确的格式 addr = ((reg << 1) & 0x7E)*/
	txdata[0] = (reg << 1) & 0x7E;
	transfer->tx_buf = txdata;
    transfer->len = 1;
	
	/* 初始化 spi_message */
	spi_message_init(&message);
	
	/* 将 spi_transger 加入 spi_message 的尾部 */
	spi_message_add_tail(transfer, &message);
	
	/* 开启 spi_sync   传输地址 */
	ret = spi_sync(rc522_spi_device, &message);
	if(ret < 0)
	{
		printk("spi_sync err!\n");
		return -1;
	}

	/* 写该寄存器的数据 */
	transfer->tx_buf = buf;
    transfer->len = len;

	/* 初始化 spi_message */
	spi_message_init(&message);

	/* 将 spi_transger 加入 spi_message 的尾部 */
	spi_message_add_tail(transfer, &message);

	/* 开启 spi_sync	 传输写该寄存器的数据 */
	ret = spi_sync(rc522_spi_device, &message);	
	if(ret < 0)
	{
		printk("spi_sync err!\n");
		return -1;
	}

	/*释放 spi_transfer */
	kfree(transfer);

	/* 片选失能 */
	gpiod_set_value(cs_gpio, 1);
	return ret;
}

/* 读一个地址的一个字节的数据*/
static unsigned char read_one_reg(unsigned char reg)
{
	unsigned char buf = 0;
	spi_read_regs(reg, &buf, 1);	
	return buf;
}

/* 写一个地址的一个字节的数据*/
static void write_one_reg(unsigned char reg, unsigned char value)
{
	unsigned char buf = value;
	spi_write_regs(reg, &buf, 1);
}

/* rc522 初始化 */
static void rc522_init(void)
{
	int ret;
	rc522_reset_disable();
	udelay(10);
	ret = gpiod_direction_output(cs_gpio, 1);
	if(ret < 0)
	{
		printk("gpio_direction_output err!\n");
	}
	udelay(10);
}
/*-------------------------------------------------------------------------*/

/* The main reason to have this class is to make mdev/udev create the
 * /dev/spidevB.C character device nodes exposing our userspace API.
 * It also simplifies memory management.
 */

static struct class *rc522_class;

/*-------------------------------------------------------------------------*/

static int spidev_probe(struct spi_device *spi)
{
	int ret = 0;
	printk("%s %s line %d\n", __FILE__, __FUNCTION__, __LINE__);
	/* 1. 记录spi设备 */

	rc522_spi_device = spi;
	
	/* 2. 注册spi字符设备 */
	major = register_chrdev(0, "100ask_rc522", &spidev_fops);
	if(major < 0)
	{
		printk("register_chrdev err!\n");
		return -1;
	}
	rc522_class = class_create(THIS_MODULE, "100ask_rc522_class");
	if(rc522_class == NULL)
	{
		printk("class_create err!\n");
		return -1;
	}
	rc522_device = device_create(rc522_class, NULL, MKDEV(major, 0), NULL, "100ask_rc522");
	if(rc522_device == NULL)
	{
		printk("device_create err!\n");
		return -1;
	}

	/* 获取cs_gpio */
	cs_gpio = gpiod_get(&spi->dev, "cs", 0);
	if(cs_gpio == NULL)
	{
		printk("gpiod_get cs_gpio err!\n");
		return -1;
	}

	/* 获取rst_gpio */
	rst_gpio = gpiod_get(&spi->dev, "rst", 0);
	if(rst_gpio == NULL)
	{
		printk("gpiod_get rst_gpio err!\n");
		return -1;
	}

	/* spi 模式0 */
    spi->mode = SPI_MODE_0;

	/* 启动 spi */
    spi_setup(spi);

	/* 记录 spi device 返回给数据传输函数 */
	rc522_spi_device = spi;

	rc522_init();
	return ret;
	
	/*rst_node = of_find_node_by_path("/rc522_rst_gpio");
	if(rst_node == NULL)
	{
		printk("of_find_node_by_path err!\n");
		return -1;
	}
	rst_gpio = of_get_named_gpio(rst_node, "rst-gpio", 0);
	if(rst_gpio < 0)
	{
		printk("of_get_named_gpio err!\n");
		return -1;
	}
	ret = gpio_direction_output(rst_gpio, 1);
	if(ret  < 0)
	{
		printk("gpiod_direction_output err!\n");
		return -1;
	}

	cs_node = of_find_node_by_path("/rc522_cs_gpio");
	if(cs_node == NULL)
	{
		printk("of_find_node_by_path err!\n");
		return -1;
	}
	cs_gpio = of_get_named_gpio(cs_node, "cs-gpio", 0);
	if(cs_gpio < 0)
	{
		printk("of_get_named_gpio err!\n");
		return -1;
	}
	ret = gpio_direction_output(cs_gpio, 1);
	if(ret  < 0)
	{
		printk("gpiod_direction_output err!\n");
		return -1;
	}
	*/
	
	/*rc522_node = of_find_node_by_path("/soc/aips-bus@02000000/spba-bus@02000000/ecspi@02008000");
	if(rc522_node == NULL)
	{
		//printk("of_find_node_by_path err!\n");
		//return -1;
	}
	cs_gpio = of_get_named_gpio(rc522_node, "cs-gpios", 0);
	if(cs_gpio < 0)
	{
		//printk("of_get_named_gpio_flags err!\n");
		//return -1;
	}
	gpio_direction_output(cs_gpio, 1);
	*/
}

static int spidev_remove(struct spi_device *spi)
{
	printk("%s %s line %d\n", __FILE__, __FUNCTION__, __LINE__);

	/* 释放 rst_gpio, cs_gpio */
	gpiod_put(rst_gpio);
	gpiod_put(cs_gpio);
	/* 卸载spi字符设备 */
	device_destroy(rc522_class, MKDEV(major, 0));
	class_destroy(rc522_class);
	unregister_chrdev(major, "100ask_rc522");
	return 0;
}


static const struct of_device_id rc522_of_match[] = {
	{ .compatible = "100ask,rc522" },
	{},
};

static const struct spi_device_id rc522_id[] = {
    {"100ask,rc522", 0},
    {}
};


static struct spi_driver rc522_spi_driver = {
	.driver = {
		.name =		"100ask_spi_rc522_drv",
		.of_match_table = of_match_ptr(rc522_of_match),
	},
	.probe =	spidev_probe,
	.remove =	spidev_remove,
	.id_table = rc522_id,
	/* NOTE:  suspend/resume methods are not necessary here.
	 * We don't do anything except pass the requests to/from
	 * the underlying controller.  The refrigerator handles
	 * most issues; the controller driver handles the rest.
	 */
};

/*-------------------------------------------------------------------------*/

static int __init rc522_spi_init(void)
{
	int status;
	printk("%s %s line %d\n", __FILE__, __FUNCTION__, __LINE__);
	status = spi_register_driver(&rc522_spi_driver);
	if (status < 0) 
		printk("spi_register_driver err!\n");
	else
		printk("spi_register_driver sucess!\n");
	return status;
}


static void __exit rc522_spi_exit(void)
{
	printk("%s %s line %d\n", __FILE__, __FUNCTION__, __LINE__);
	spi_unregister_driver(&rc522_spi_driver);
}

module_init(rc522_spi_init);
module_exit(rc522_spi_exit);
MODULE_LICENSE("GPL");
MODULE_AUTHOR("JunCxue"); 

